/*
 * bswImpl.c
 *
 * implementation for shared library
 *
 * This wizard-generated code is based on code adapted from the
 * SampleLib project distributed as part of the Palm OS SDK 4.0.
 *
 * Copyright (c) 1994-1999 Palm, Inc. or its subsidiaries.
 * All rights reserved.
 */

/* Our library public definitions (library API) */
#define BUILDING_BSW
#include "bsw.h"
#include "bswPrivate.h"

/*********************************************************************
 *
 * LIBRARY GLOBALS:
 *
 * IMPORTANT:
 * ==========
 * Libraries are *not* allowed to have global or static variables.
 * Instead, they allocate a memory chunk to hold their persistent data, 
 * and save a handle to it in the library's system library table entry.
 * Example functions below demostrate how the library "globals" chunk 
 * is set up, saved, and accessed.
 *
 * We use a movable memory chunk for our library globals to minimize
 * dynamic heap fragmentation.  Our library globals are locked only
 * when needed and unlocked as soon as possible.  This behavior is
 * critical for healthy system performance.
 *
 *********************************************************************/


/*********************************************************************
 * Internally used routines
 *********************************************************************/

static bswGlobalsType * PrvMakeGlobals(UInt16 refNum);
static void PrvFreeGlobals(UInt16 refNum);
static bswGlobalsType * PrvLockGlobals(UInt16 refNum);
static Boolean PrvIsLibOpen(UInt16 refNum);
static Err PrvCreateClientContext(bswGlobalsType * gP, UInt32 * clientContextP);
static Err PrvDestroyClientContext(bswGlobalsType * gP, UInt32 clientContext);
static bswClientContextType * PrvLockContext(UInt32 context);

static Err PrvOpenScheme(UInt16 refNum, UInt32 clientContext, UInt32 schemeCode);
static Err PrvImplementationCall(UInt16 refNum, MemPtr cmdPBP, UInt16 launchCode);
static Err PrvGetImplementation(UInt16 refNum, bswLaunchCode launchCode, UInt16 *cardNoP, LocalID *dbIDP);
static Err PrvGetBooksInt16Resource(UInt32 recordType, UInt16 bookID, UInt16 *resultP);
static Err PrvGetChaptersInt16Resource(UInt32 recordType, UInt16 bookID, UInt16 chapterNum, UInt16 *resultP);
static Err PrvGetInt16Resource(UInt32 recordType, UInt16 recordID, UInt16 elementNumber, UInt16 *resultP);

/*********************************************************************
 * Internally used macros
 *********************************************************************/

/* Unlock globals */
#define PrvUnlockGlobals(gP)        MemPtrUnlock(gP)

/* Unlock the client context */
#define PrvUnlockContext(contextP)	MemPtrUnlock(contextP)
 
/* LIBRARY VERSION
 *
 * The library version scheme follows the system versioning scheme.
 * See sysMakeROMVersion and friends in SystemMgr.h.
 *
 * For reference:
 *
 * 0xMMmfsbbb, where MM is major version, m is minor version
 * f is bug fix, s is stage: 3-release,2-beta,1-alpha,0-development,
 * bbb is build number for non-releases 
 * V1.12b3   would be: 0x01122003
 * V2.00a2   would be: 0x02001002
 * V1.01     would be: 0x01013000
 */

#define prvLibVersion	sysMakeROMVersion(1, 0, 0, sysROMStageDevelopment, 1)

/*********************************************************************
 * Library API Routines
 *********************************************************************/
 
/*
 * FUNCTION: bswOpen
 *
 * DESCRIPTION:
 *
 * Opens the library, creates and initializes the globals.
 * This function must be called before any other library functions.
 *
 * If bswOpen fails, do not call any other library API functions.
 * If bswOpen succeeds, call bswClose when you are done using
 * the library to enable it to release critical system resources.
 *
 * LIBRARY DEVELOPER NOTES:
 *
 * The library's "open" and "close" functions should *not* take an excessive
 * amount of time to complete.  If the processing time for either of these
 * is lengthy, consider creating additional library API function(s) to handle
 * the time-consuming chores.
 *
 *
 * PARAMETERS:
 *
 * refNum
 *		Library reference number returned by SysLibLoad() or SysLibFind().
 *
 * clientContextP
 *		pointer to variable for returning client context.  The client context is
 *		used to maintain client-specific data for multiple client support.  The 
 *		value returned here will be used as a parameter for other library 
 *		functions which require a client context.  
 *
 * CALLED BY: anyone who wants to use this library
 *
 * RETURNS:
 *		errNone
 *		memErrNotEnoughSpace
 *
 * SIDE EFFECTS:
 *		*clientContextP will be set to client context on success, or zero on error.
 */

Err bswOpen(UInt16 refNum, UInt32 *clientContextP)
{
	bswGlobalsType *gP;
	Err err = errNone;
	Int16 originalOpenCount = 0;

	/* Error-check our parameters */
	ErrFatalDisplayIf(
		clientContextP == NULL, 
		"null context variable pointer");

	/* Initialize return variable */
	*clientContextP = 0;

	/* Get library globals */
	gP = PrvLockGlobals(refNum);

	/* Check if already open */
	if (!gP)
	{
		/* Allocate and initialize our library globals. */
		gP = PrvMakeGlobals(refNum);
		if ( !gP )
			err = memErrNotEnoughSpace;
	}

	/* If we have globals, create a client context, increment open
	 * count, and unlock our globals */
	if ( gP )
	{
		originalOpenCount = gP->openCount;

		err = PrvCreateClientContext(gP, clientContextP);
		if ( !err )
			gP->openCount++;

		PrvUnlockGlobals(gP);

		/* If there was an error creating a client context and there  */
		/* are no other clients, free our globals */
		if ( err && (originalOpenCount == 0) )
			PrvFreeGlobals(refNum);
	}

	return( err );
}

/*
 * FUNCTION: bswClose
 *
 * DESCRIPTION:	
 *
 * Closes the library, frees client context and globals.
 *
 * ***IMPORTANT***
 * May be called only if bswOpen succeeded.
 *
 * If other applications still have the library open, decrements
 * the reference count and returns bswErrStillOpen.
 *
 * LIBRARY DEVELOPER NOTES:
 *
 * The library's "open" and "close" functions should *not* take an excessive
 * amount of time to complete.  If the processing time for either of these
 * is lengthy, consider creating additional library API function(s) to handle
 * the time-consuming chores.
 *
 * PARAMETERS:
 *
 * refNum
 *		Library reference number returned by SysLibLoad() or SysLibFind().
 *
 * clientContext
 *		client context
 *
 * CALLED BY: Whoever wants to close the library
 *
 * RETURNS:
 *		errNone
 *		bswErrStillOpen -- library is still open by others (no error)
 */

Err bswClose(UInt16 refNum, UInt32 clientContext)
{
	bswGlobalsType * gP;
	Int16 openCount;
	Int16 contextCount;
	Err err = errNone;

	gP = PrvLockGlobals(refNum);

	/* If not open, return */
	if (!gP)
	{
		/* MUST return zero here to get around a bug in system v1.x that
		 * would cause SysLibRemove to fail. */
		return errNone;
	}

	/* Destroy the client context (we ignore the return code in this implementation) */
	PrvDestroyClientContext(gP, clientContext);

	/* Decrement our library open count */
	gP->openCount--;

	/* Error check for open count underflow */
	ErrFatalDisplayIf(gP->openCount < 0, "BibleSwitchLibrary open count underflow");

	/* Save the new open count and the context count */
	openCount = gP->openCount;
	contextCount = gP->contextCount;

	PrvUnlockGlobals(gP);

	/* If open count reached zero, free our library globals */
	if ( openCount <= 0 )
	{
		/* Error check to make sure that all client contexts were destroyed */
		ErrFatalDisplayIf(contextCount != 0, "not all client contexts were destroyed");

		/* Free our library globals */
		PrvFreeGlobals(refNum);
	}
	else
	{
		/* return this error code to inform the caller
		 * that others are still using this library */
		err = bswErrStillOpen;
	}

	return err;
}

/*
 * FUNCTION: bswSleep
 *
 * DESCRIPTION:
 *
 * Handles system sleep notification.
 *
 * ***IMPORTANT***
 * This notification function is called from a system interrupt.
 * It is only allowed to use system services which are interrupt-
 * safe.  Presently, this is limited to EvtEnqueueKey, SysDisableInts,
 * SysRestoreStatus.  Because it is called from an interrupt,
 * it must *not* take a long time to complete to preserve system
 * integrity.  The intention is to allow system-level libraries
 * to disable hardware components to conserve power while the system
 * is asleep.
 * 
 * PARAMETERS:
 *
 * refNum
 *		Library reference number returned by SysLibLoad() or SysLibFind().
 *
 * CALLED BY: System
 *
 * RETURNS:	errNone
 */

Err bswSleep(UInt16 refNum)
{
	#pragma unused(refNum)
	return errNone;
}

/*
 * FUNCTION: bswWake
 *
 * DESCRIPTION:
 *
 * Handles system wake notification.
 *
 * ***IMPORTANT***
 * This notification function is called from a system interrupt.
 * It is only allowed to use system services which are interrupt-
 * safe.  Presently, this is limited to EvtEnqueueKey, SysDisableInts,
 * SysRestoreStatus.  Because it is called from an interrupt,
 * it must *not* take a long time to complete to preserve system
 * integrity.  The intention is to allow system-level libraries
 * to enable hardware components which were disabled when the system
 * went to sleep.
 * 
 * PARAMETERS:
 *
 * refNum
 *		Library reference number returned by SysLibLoad() or SysLibFind().
 *
 * CALLED BY: System
 *
 * RETURNS:	errNone
 */

Err bswWake(UInt16 refNum)
{
	#pragma unused(refNum)
	return errNone;
}

/************************************************************
 *
 *  FUNCTION: bswVersion
 *
 *  DESCRIPTION:	Get our library API version.  The Sample library does not
 *					need to be "opened" to call bswVersion.
 *
 *  PARAMETERS:	refNum		-- Sample library reference number
 *				verP		-- pointer to variable for storing the version number
 *
 *  CALLED BY:		Anyone wishing to get our library API version
 *
 *  RETURNS:	errNone		-- no error
 *				bswErrParam	-- if verP is NULL
 *
 *  CREATED: 5/20/97 
 *
 *  BY: vmk
 *
 *  REVISION HISTORY:
 *			Name	Date		Description
 *			----	----		-----------
 *			SPD		2003-11-21	Copied from SDK examples SampleLib.c SampleLibGetLibAPIVersion
 *			SPD		2003-11-21	Changed from using ErrFatalDisplayIf to returning bswErrParam
 *
 */
Err bswVersion(UInt16 refNum, UInt32 * verP)
{
	/* Error-check our parameters */
	if( verP == NULL) return bswErrParam;
	*verP = prvLibVersion;
	return( errNone );
}

/************************************************************
 *
 *  FUNCTION: bswHasImplementation
 *
 *  DESCRIPTION:	Find out if the launch code has had an implementation specified, and that the implementor exists.
 *
 *  PARAMETERS:
 *
 *	refNum
 *		Sample library reference number
 *
 *	launchCode
 *		implementable function to test for the existance of an implementor
 *
 *  CALLED BY:	Anyone wishing to test the capabilities of the current setup.
 *
 *  RETURNS:
 *		bswErrParam if launchCode is NULL.
 *		dmErrCantFind if the specified implementing application can not be found.
 *		errNone if the specified implementing application can be found.
 *
 *  REVISION HISTORY:
 *			Name	Date		Description
 *			----	----		-----------
 *			DT		2004-02-18	Raised need for routine
 *			SPD		2004-03-03	Created routine
 *
 */
Err bswHasImplementation(UInt16 refNum, bswLaunchCode launchCode)
{
	UInt16 cardNo;
	LocalID dbID;
	Err err = errNone;
	
	err = PrvGetImplementation(refNum, launchCode, &cardNo, &dbID);
	//ignore cardNo & dbID. If not found, an error is returned.
    return err;
}


/************************************************************
 * FUNCTION: bswLaunch
 *
 * DESCRIPTION:
 *
 * Switch to the prefered bible application, viewing the supplied reference.
 *
 * TODO: If the preference has not been set, or is set to an application that can not be found
 * then search for implementing applications? Select the first? Set new preferences value?
 * 
 * PARAMETERS:
 *
 * refNum
 *		Library reference number returned by SysLibLoad() or SysLibFind().
 *
 * refP
 *      Pointer to reference structure containing the verse to view
 *
 * CALLED BY: User
 *
 * RETURNS:
 *		bswErrParam	if refP is NULL
 *      dmErrCantFind if the specified implementing application can not be found.
 *		memErrNotEnoughSpace if there is not enough space to allocate the launch command parameter block.
 *		memErrInvalidParam is a return code MemPtrSetOwner is able to give, but shouldn't.
 *		errNone
 */
Err bswLaunch(UInt16 refNum, bswReference *refP)
{
	/* Error-check our parameters */
	if( refP == NULL) return bswErrParam;

	/* Version-check our parameters */
	if( refP->structureSize != bswReferenceStructureSizeV1 ) return bswErrBadStructureSize;

	return PrvImplementationCall(refNum, (MemPtr)refP, bswAppLaunchCmdLaunch);
}

/************************************************************
 * FUNCTION: bswGetReference
 *
 * DESCRIPTION:
 *
 * Retrieve the text for the reference
 * 
 * PARAMETERS:
 *
 * refNum
 *		Library reference number returned by SysLibLoad() or SysLibFind().
 *
 * refP
 *		Pointer to structure containing Bible Switchboard Reference
 *
 * CALLED BY: User
 *
 * RETURNS:
 *		errNone
 */
Err bswGetReference(UInt16 refNum, bswText *textP)
{
	/* Error-check our parameters */
	if( textP == NULL) return bswErrParam;

	/* Version-check our parameters */
	if( textP->structureSize != bswTextStructureSizeV1 ) return bswErrBadStructureSize;

	return PrvImplementationCall(refNum, (MemPtr)textP, bswAppLaunchCmdGetReference);
}    

/************************************************************
 * FUNCTION: bswFindText
 *
 * DESCRIPTION:
 *
 * Search the bible for the given text, returning a list of verses (bitmap?)
 * 
 * PARAMETERS:
 *
 * refNum
 *		Library reference number returned by SysLibLoad() or SysLibFind().
 *
 * refP
 *		Pointer to search structure
 *
 * CALLED BY: User
 *
 * RETURNS:
 *		errNone
 */
Err bswFindText(UInt16 refNum, bswSearch *searchP)
{
	/* Error-check our parameters */
	if( searchP == NULL) return bswErrParam;

	/* Version-check our parameters */
	if( searchP->structureSize != bswSearchStructureSizeV1 ) return bswErrBadStructureSize;

	return PrvImplementationCall(refNum, (MemPtr)searchP, bswAppLaunchCmdFindText);
}    

/************************************************************
 * FUNCTION: bswSelectReference
 *
 * DESCRIPTION:
 *
 * Open the preferred verse selector screen, returning the selected reference
 * 
 * PARAMETERS:
 *
 * refNum
 *		Library reference number returned by SysLibLoad() or SysLibFind().
 *
 * refP
 *		Pointer to structure containing Bible Switchboard Reference
 *
 * CALLED BY: User
 *
 * RETURNS:
 *		errNone
 */
Err bswSelectReference(UInt16 refNum, bswReference *refP)
{
	/* Error-check our parameters */
	if( refP == NULL) return bswErrParam;

	/* Version-check our parameters */
	if( refP->structureSize != bswReferenceStructureSizeV1 ) return bswErrBadStructureSize;

	return PrvImplementationCall(refNum, (MemPtr)refP, bswAppLaunchCmdSelectReference);
}    

/************************************************************
 * FUNCTION: bswSelectedReference
 *
 * DESCRIPTION:
 *
 * Get the reference currently in use by the preferred bible application
 * 
 * PARAMETERS:
 *
 * refNum
 *		Library reference number returned by SysLibLoad() or SysLibFind().
 *
 * refP
 *		Pointer to structure containing Bible Switchboard Reference to fill in
 *
 * CALLED BY: User
 *
 * RETURNS:
 *      dmErrCantFind if the specified implementing application can not be found.
 *		errNone
 */

Err bswSelectedReference(UInt16 refNum, bswReference *refP)
{
	/* Error-check our parameters */
	if( refP == NULL) return bswErrParam;

	/* Version-check our parameters */
	if( refP->structureSize != bswReferenceStructureSizeV1 ) return bswErrBadStructureSize;

	return PrvImplementationCall(refNum, (MemPtr)refP, bswAppLaunchCmdSelectedReference);
}    

Err bswGetCount(UInt16 refNum, UInt32 clientContext, bswCount *countP)
{
	/* Error-check our parameters */
	if( countP == NULL) return bswErrParam;

	/* Version-check our parameters */
	if( countP->structureSize != bswCountStructureSizeV1 ) return bswErrBadStructureSize;
	if( countP->withinContextReference != NULL ) {
		if( countP->withinContextReference->structureSize != bswReferenceStructureSizeV1 ) return bswErrBadStructureSize;
	}

	return PrvImplementationCall(refNum, (MemPtr)countP, bswAppLaunchCmdGetCount);
}

Err bswGetTranslationList(UInt16 refNum, UInt32 clientContext, UInt16 whereFrom, bswTranslationList **translationListP)
{
	UInt16 cardNo;
	LocalID dbID;
	Err err = errNone;
	UInt32 result;
	bswClientContextTypePtr contextP;
	bswTranslationList * tLP;
	UInt16 i;
	UInt32 thisImplementorID;
	UInt16 prefsSize;
	UInt16 thisFunc = whereFrom - bswAppLaunchCmdLaunch;
	
	/* Error-check our parameters */
	if( whereFrom == 0 ) return bswErrParam;
	if( translationListP == NULL) return bswErrParam;
	if ( !PrvIsLibOpen(refNum) )
		return( bswErrNotOpen );

	contextP = PrvLockContext(clientContext);
	tLP = contextP->tListP[thisFunc].translations;

	if (tLP == NULL) {
		// Has this implementor been asked for a list before, that covers this new function?
		prefsSize=4;
		thisImplementorID=NULL;
		PrefGetAppPreferences(bswPreferencesCreatorID,thisFunc + 1,&thisImplementorID,&prefsSize,true);
		contextP->tListP[thisFunc].implementorID = thisImplementorID;
		i = 0;
		while ((tLP == NULL) && (i < 32)) {
			if (i!=thisFunc) {
				if (contextP->tListP[i].implementorID == thisImplementorID) {
					if (contextP->tListP[i].translations->forAPI & (1<<thisFunc) == (1<<thisFunc)) {
						//found one cached
						tLP = contextP->tListP[i].translations;
						contextP->tListP[thisFunc].translations = tLP;
					}
				}
			}
			i++;
		}
		if (tLP == NULL) {
			//none cached yet, ask the implementor
			tLP = (bswTranslationList *) MemPtrNew(sizeof(bswTranslationList));
			tLP->structureSize = sizeof(bswTranslationList);
			tLP->numTranslations = 0;
			tLP->forAPI = 0;
			tLP->translationArray = NULL;
			tLP->translationNames = NULL;
			tLP->translationCodes = NULL;
			contextP->tListP[thisFunc].translations = tLP;
			err = PrvGetImplementation(refNum, whereFrom, &cardNo, &dbID);
			if( errNone == err ) {
				err = SysAppLaunch(cardNo, dbID, 0, bswAppLaunchCmdGetTranslationList, (MemPtr) tLP, &result);
				if (errNone == err) {
					err = result;
				}
			}
		}
	}
	
	*translationListP = tLP;

	PrvUnlockContext(contextP);

    return err;
}

Err bswFreeTranslationList(UInt16 refNum, UInt32 clientContext, UInt16 whereFrom, bswTranslationList **translationListP)
{
	UInt16 cardNo;
	LocalID dbID;
	Err err = errNone;
	UInt32 result;
	
	/* Error-check our parameters */
	if( whereFrom == 0 ) return bswErrParam;
	if( translationListP == NULL) return bswErrParam;
	if( *translationListP == NULL) return errNone;

	/* Version-check our parameters */
	if( (*translationListP)->structureSize != bswTranslationListStructureSizeV1 ) return bswErrBadStructureSize;

	err = PrvGetImplementation(refNum, whereFrom, &cardNo, &dbID);
	if( errNone == err ) {
		err = SysAppLaunch(cardNo, dbID, 0, bswAppLaunchCmdFreeTranslationList, (MemPtr) *translationListP, &result);
		if (errNone == err) {
			err = result;
			MemPtrFree(*translationListP);
			*translationListP = NULL;
		}
	}

    return err;
}


/************************************************************
 * FUNCTION: bswGetBookID
 *
 * DESCRIPTION:
 *
 * ?
 * 
 * PARAMETERS:
 *
 * refNum
 *		Library reference number returned by SysLibLoad() or SysLibFind().
 *
 * CALLED BY: User
 *
 * RETURNS:
 *		errNone
 */

Err bswGetBookID(UInt16 refNum, UInt32 clientContext, UInt16 seqNum, UInt16 *bookIDP)
{
	/* Error-check our parameters */
	if( bookIDP == NULL) return bswErrParam;

	PrvOpenScheme(refNum, clientContext, 0);

	return PrvGetInt16Resource('BNSq', 0, seqNum-1, bookIDP);
}

/************************************************************
 * FUNCTION: bswGetBooksSeqNum
 *
 * DESCRIPTION:
 *
 * ?
 * 
 * PARAMETERS:
 *
 * refNum
 *		Library reference number returned by SysLibLoad() or SysLibFind().
 *
 * CALLED BY: User
 *
 * RETURNS:
 *		errNone
 */

Err bswGetBooksSeqNum(UInt16 refNum, UInt32 clientContext, UInt16 bookID, UInt16 *seqNumP)
{
	/* Error-check our parameters */
	if( seqNumP == NULL) return bswErrParam;

	PrvOpenScheme(refNum, clientContext, 0);

	return PrvGetInt16Resource('BOrd', bookID, 0, seqNumP);
}

/************************************************************
 * FUNCTION: bswGetBookNamesForBook
 *
 * DESCRIPTION:
 *
 * ?
 * 
 * PARAMETERS:
 *
 * refNum
 *		Library reference number returned by SysLibLoad() or SysLibFind().
 *
 * CALLED BY: User
 *
 * RETURNS:
 *		errNone
 *		bswErrLock
 *		bswErrUnknownBook
 */

Err bswGetBookNames(UInt16 refNum, UInt32 clientContext, bswReferenceType typeFlag, Char ***allNamesArrayP, MemHandle *namesResourceP)
{
	MemHandle h = NULL;
	UInt8* p = NULL;
	UInt16 numStrings = 0;
	MemHandle hS = NULL;
	char** pS = NULL;
	
	/* Error-check our parameters */
	if( allNamesArrayP == NULL) return bswErrParam;
	if( namesResourceP == NULL) return bswErrParam;

	PrvOpenScheme(refNum, clientContext, 0);

	switch (typeFlag) {
		case bswTextReferenceType:
			h=DmGetResource('tSTL',1);
			break;
		case bswTinyTextReferenceType:
			h=DmGetResource('tSTL',2);
			break;
		case bswShortTextReferenceType:
			h=DmGetResource('tSTL',3);
			break;
		default:
			return bswErrParam;
	}

	if (!h) {
		return bswErrResource;
	}

	p=(UInt8*)MemHandleLock(h);
	if (!p) {
	    DmReleaseResource(h);
		return bswErrLock;
	}

	numStrings=p[2];

	hS=SysFormPointerArrayToStrings((char*)&p[3],numStrings);
	if(!hS){
	    MemHandleUnlock(h);
	    DmReleaseResource(h);
		return bswErrLock;
	}

	pS=(char**)MemHandleLock(hS);
	if(!pS){
	    MemHandleFree(hS);
	    MemHandleUnlock(h);
	    DmReleaseResource(h);
		return bswErrLock;
	}
	
	*allNamesArrayP = pS;
	*namesResourceP = h;

    return errNone;
}

/************************************************************
 * FUNCTION: bswGetBookNamesForBook
 *
 * DESCRIPTION:
 *
 * ?
 * 
 * PARAMETERS:
 *
 * refNum
 *		Library reference number returned by SysLibLoad() or SysLibFind().
 *
 * CALLED BY: User
 *
 * RETURNS:
 *		errNone
 *		bswErrLock
 *		bswErrUnknownBook
 */

Err bswGetBooksNames(UInt16 refNum, UInt32 clientContext, UInt16 bookID, UInt16 *numNamesP, Char ***allNamesArrayP, MemHandle *namesResourceP)
{
	MemHandle h = NULL;
	UInt8* p = NULL;
	UInt16 numStrings = 0;
	MemHandle hS = NULL;
	char** pS = NULL;
	
	/* Error-check our parameters */
	if( numNamesP == NULL) return bswErrParam;
	if( allNamesArrayP == NULL) return bswErrParam;
	if( namesResourceP == NULL) return bswErrParam;

	PrvOpenScheme(refNum, clientContext, 0);

	h=DmGetResource('tSTL',bookID);
	if (!h) {
		return bswErrUnknownBook;
	}

	p=(UInt8*)MemHandleLock(h);
	if (!p) {
	    DmReleaseResource(h);
		return bswErrLock;
	}

	numStrings=p[2];

	hS=SysFormPointerArrayToStrings((char*)&p[3],numStrings);
	if(!hS){
	    MemHandleUnlock(h);
	    DmReleaseResource(h);
		return bswErrLock;
	}

	pS=(char**)MemHandleLock(hS);
	if(!pS){
	    MemHandleFree(hS);
	    MemHandleUnlock(h);
	    DmReleaseResource(h);
		return bswErrLock;
	}
	
	*numNamesP = numStrings;
	*allNamesArrayP = pS;
	*namesResourceP = h;

    return errNone;
}

/************************************************************
 * FUNCTION: bswFreeBookNames
 *
 * DESCRIPTION:
 *
 * ?
 * 
 * PARAMETERS:
 *
 * refNum
 *		Library reference number returned by SysLibLoad() or SysLibFind().
 *
 * CALLED BY: User
 *
 * RETURNS:
 *		errNone
 */

Err bswFreeBookNames(UInt16 refNum, UInt32 clientContext, Char ***allNamesArrayP, MemHandle *namesResourceP)
{
	/* Error-check our parameters */
	/* commented out as this routine behaves safely anyway
	if( allNamesArrayP == NULL) return bswErrParam;
	if( namesResourceP == NULL) return bswErrParam;
	*/

	if (allNamesArrayP) {
		if (*allNamesArrayP) {
		    MemPtrUnlock(*allNamesArrayP);
		    MemPtrFree(*allNamesArrayP);
		    *allNamesArrayP = NULL;
		}
	}
	if (namesResourceP) {
		if (*namesResourceP) {
		    MemHandleUnlock(*namesResourceP);
		    DmReleaseResource(*namesResourceP);
		    *namesResourceP = NULL;
		}
	}
    return errNone;
}

/************************************************************
 * FUNCTION: bswCompleteRef
 *
 * DESCRIPTION:
 *
 *		Translates between standard, text, and numeric references, and sanity checks the results.
 *
 *		If completeFlags has bswTextReferenceType set, the returned string will be newly allocated and will need to be freed by the user.
 *		The user may want to clean up a text reference, by setting useFlag == completeFlags == bswTextReferenceType.
 *		This will overwrite the existing text reference pointer, so the user should save the old value if it is needed.
 * 
 * PARAMETERS:
 *
 * refNum
 *		Library reference number returned by SysLibLoad() or SysLibFind().
 *
 * refP
 *		Pointer to the reference used for both the source of the reference, and the destination of the translation.
 *
 * useFlag
 *		one of bswStandardReferenceType, bswTextReferenceType, or bswNumericReferenceType
 *
 * completeFlags
 *		combination of bswStandardReferenceType, bswTextReferenceType, or bswNumericReferenceType
 *		specifying which parts of the reference to fill in.
 *
 * CALLED BY: User
 *
 * RETURNS:
 *		errNone
 */

Err bswCompleteRef(UInt16 refNum, UInt32 clientContext, bswReference *refP, bswReferenceType useFlag, bswReferenceType completeFlags)
{
	bswReference tmpRef;
	Err err = errNone;
	MemHandle h = NULL;
	UInt16* p = NULL;
	MemHandle hS = NULL;
	char** pS = NULL;
	Char* sBookName = NULL;
	Char sChapterNumber[maxStrIToALen+1];
	Char sVerseNumber[maxStrIToALen+1];
	UInt16 bn = 0;
	UInt16 cn = 0;
	UInt16 vn = 0;
	UInt16 bfcn = 0;
	UInt8* p8 = NULL;
	UInt16 numStrings = 0;
	UInt16 sLen = 0;
    Boolean match=false;
    UInt16 fieldNum;


	/* Error-check our parameters */
	if( refP == NULL) return bswErrParam;

	/* Version-check our parameters */
	if( refP->structureSize != bswReferenceStructureSizeV1 ) return bswErrBadStructureSize;

	MemSet(&tmpRef,sizeof(bswReference),0);

	//PrvOpenScheme(refNum, clientContext, refP->schemeRef);
	PrvOpenScheme(refNum, clientContext, 0);

	switch (useFlag) {
		case bswStandardReferenceType:
			if( refP->bookNumber == 0 ) return bswErrParam;

			tmpRef.bookNumber = refP->bookNumber;
			tmpRef.chapterNumber = refP->chapterNumber;
			tmpRef.verseNumber = refP->verseNumber;

			break;
		case bswTextReferenceType:
			if( refP->textRef == NULL ) return bswErrParam;
			
			/* split into bookname, chapter, verse
			 
			 digits are allowed to be part of the book name until a non digit character,
			 then digits signify the start of the chapter
			 so search for first non digit, then first digit (none means no chapter, verse reference)
			 then for non digit/EOS (end of chapter number),
			 then for digit (optional verse number)
			 then for non digit/EOS (end of verse number),

			 if trimmed chapter/verse separator is "-" it is chapter range, 
			 eg The Story of Noah is in "Genesis 6-9"
			 
			 */
			
			bn = 0;
			while (refP->textRef[bn]==' ') bn++; /* Trim starting spaces */
			tmpRef.textRef=&(refP->textRef[bn]);
			bn=0;
			while ((tmpRef.textRef[bn]!=0)&&(tmpRef.textRef[bn]>='0')&&(tmpRef.textRef[bn]<='9')) bn++; /* accept all digits */
			while ((tmpRef.textRef[bn]!=0)&&((tmpRef.textRef[bn]<'0')||(tmpRef.textRef[bn]>'9'))) bn++; /* accept all non-digits */
			/* bn is either at a digit or end of string */
			cn=bn;
			if (bn>0) bn--;
			while ((bn>0)&&(tmpRef.textRef[bn]==' ')) bn--; /* Trim trailing spaces */
			sBookName=MemPtrNew(bn+2);
			StrNCopy(sBookName,tmpRef.textRef,bn+1);
			sBookName[bn+1]=0;

			tmpRef.chapterNumber = 0;
			while ((tmpRef.textRef[cn]!=0)&&(tmpRef.textRef[cn]>='0')&&(tmpRef.textRef[cn]<='9')) {
				tmpRef.chapterNumber = tmpRef.chapterNumber * 10 + (tmpRef.textRef[cn] - '0');
				cn++;
			}
			/* cn is either at a non-digit or end of string */
			while ((tmpRef.textRef[cn]!=0)&&(tmpRef.textRef[cn]!='-')&&((tmpRef.textRef[cn]<'0')||(tmpRef.textRef[cn]>'9'))) cn++; /* accept all non-digits except hyphen */
			vn=cn;
			tmpRef.verseNumber = 0;
			while ((tmpRef.textRef[vn]!=0)&&(tmpRef.textRef[vn]>='0')&&(tmpRef.textRef[vn]<='9')) {
				tmpRef.verseNumber = tmpRef.verseNumber * 10 + (tmpRef.textRef[vn] - '0');
				vn++;
			}

			err = PrvGetInt16Resource('TNBs', 0, 0, &bn);

			while ( (bn > 0) & !match ) {

				err = PrvGetInt16Resource('BNSq', 0, bn-1, &tmpRef.bookNumber);
				
				h=DmGetResource('tSTL',tmpRef.bookNumber);
				if (!h) {
					return bswErrResource;
				}

				p8=(UInt8*)MemHandleLock(h);
				if (!p8) {
				    DmReleaseResource(h);
					return bswErrLock;
				}

				numStrings=p8[2];

				hS=SysFormPointerArrayToStrings((char*)&p8[3],numStrings);
				if(!hS){
				    MemHandleUnlock(h);
				    DmReleaseResource(h);
					return bswErrLock;
				}

				pS=(char**)MemHandleLock(hS);
				if(!pS){
				    MemHandleFree(hS);
				    MemHandleUnlock(h);
				    DmReleaseResource(h);
					return bswErrLock;
				}
				
				fieldNum=0;
				while ( !match && (fieldNum < numStrings) ){
		            if (StrCaselessCompare(pS[fieldNum],sBookName)==0){
		                match = true;
		            } else {
		                fieldNum++;
		            }
		        }
			    MemHandleUnlock(hS);
			    MemHandleFree(hS);
			    MemHandleUnlock(h);
			    DmReleaseResource(h);
			    bn--;
			}
			
			MemPtrFree(sBookName);
			
			if (!match) {
				tmpRef.bookNumber=0;
			}

			break;
		case bswNumericReferenceType:
			if( refP->verseRef == 0 ) return bswErrParam;
			
			err = PrvGetInt16Resource('TNBs', 0, 0, &bn);
			h=DmGetResource('BFVN',0); if (!h) return bswErrResource;
			p=(UInt16*)MemHandleLock(h); if (!p) { DmReleaseResource(h); return bswErrLock; }
			while (refP->verseRef<p[bn-1]){
				bn--;
			}
		    MemHandleUnlock(h);
		    DmReleaseResource(h);

			err = PrvGetInt16Resource('BFCN', 0, bn-1, &bfcn);
			err = PrvGetInt16Resource('BNCs', 0, bn-1, &(tmpRef.chapterNumber));
			h=DmGetResource('CFVN',0); if (!h) return bswErrResource;
			p=(UInt16*)MemHandleLock(h); if (!p) { DmReleaseResource(h); return bswErrLock; }
			while (refP->verseRef<p[bfcn-1 + tmpRef.chapterNumber-1]){
				tmpRef.chapterNumber--;
			}
		    MemHandleUnlock(h);
		    DmReleaseResource(h);

			tmpRef.verseNumber = refP->verseRef - p[bfcn-1 + tmpRef.chapterNumber-1] + 1;
			
			err = PrvGetInt16Resource('BNSq', 0, bn-1, &(tmpRef.bookNumber));
			
			break;
		default:
			return bswErrParam;
	}

	/* Check tmpRef, restrict to maximum chapter/verse if chapter,verse numbers are too high. */
	
	err = PrvGetInt16Resource('BOrd', tmpRef.bookNumber, 0, &bn);
	if ( err != errNone ) return bswErrUnknownBook;

	err = PrvGetInt16Resource('BNCs', 0, bn-1, &cn);
	if (cn < tmpRef.chapterNumber) {
		tmpRef.chapterNumber = cn;
		if (tmpRef.verseNumber > 0) {
			/* last verse in last chapter? Will get corrected below */
			tmpRef.verseNumber = -1;
		}
	}
	
	err = PrvGetInt16Resource('BFCN', 0, bn-1, &bfcn);

	h=DmGetResource('CNVs',0); if (!h) return bswErrResource;
	p=(UInt16*)MemHandleLock(h); if (!p) { DmReleaseResource(h); return bswErrLock; }
	vn=p[bfcn-1 + tmpRef.chapterNumber-1];
	if (vn < tmpRef.verseNumber) {
		tmpRef.verseNumber = vn;
	}
    MemHandleUnlock(h);
    DmReleaseResource(h);

	if ((completeFlags & bswStandardReferenceType) != 0) {
		refP->bookNumber = tmpRef.bookNumber;
		refP->chapterNumber = tmpRef.chapterNumber;
		refP->verseNumber = tmpRef.verseNumber;
	}

	if ((completeFlags & (bswTextReferenceType | bswTinyTextReferenceType | bswShortTextReferenceType)) != 0) {
		/* Calculate string length */
		
		h=DmGetResource('tSTL',tmpRef.bookNumber);
		if (!h) {
			return bswErrResource;
		}

		p8=(UInt8*)MemHandleLock(h);
		if (!p8) {
		    DmReleaseResource(h);
			return bswErrLock;
		}

		numStrings=p8[2];

		hS=SysFormPointerArrayToStrings((char*)&p8[3],numStrings);
		if(!hS){
		    MemHandleUnlock(h);
		    DmReleaseResource(h);
			return bswErrLock;
		}

		pS=(char**)MemHandleLock(hS);
		if(!pS){
		    MemHandleFree(hS);
		    MemHandleUnlock(h);
		    DmReleaseResource(h);
			return bswErrLock;
		}

		switch (completeFlags & (bswTextReferenceType | bswTinyTextReferenceType | bswShortTextReferenceType)) {
			case bswTextReferenceType:
				sBookName = pS[0];
				break;
			case bswTinyTextReferenceType:
				sBookName = pS[1];
				break;
			case bswShortTextReferenceType:
				sBookName = pS[2];
				break;
		}
		sLen = StrLen(sBookName);
		if (tmpRef.chapterNumber > 0) {
			sChapterNumber[0]=' ';
			StrIToA(&(sChapterNumber[1]), tmpRef.chapterNumber);
			sLen += StrLen(sChapterNumber);
			if (tmpRef.verseNumber > 0) {
				sVerseNumber[0]=':';
				StrIToA(&(sVerseNumber[1]), tmpRef.verseNumber);
				sLen += StrLen(sVerseNumber);
			}
		}
		sLen++; /* Enf of String byte */

		refP->textRef = MemPtrNew(sLen);
		if ( refP->textRef != NULL ) {
			StrCopy(refP->textRef,sBookName);
			if (tmpRef.chapterNumber > 0) {
				StrCat(refP->textRef,sChapterNumber);
				if (tmpRef.verseNumber > 0) {
					StrCat(refP->textRef,sVerseNumber);
				}
			}
		}
		
	    MemHandleUnlock(hS);
	    MemHandleFree(hS);
	    MemHandleUnlock(h);
	    DmReleaseResource(h);
	}

	if ((completeFlags & bswNumericReferenceType) != 0) {
		//My book numbering is as per http://tirith.mine.nu/jsp/BibleConverterBookNumbers.html
		//this routine converts from book numbers (from 10 to 730) to ordinal positions, 1-based index (from 1 to 66)
		//err = PrvGetInt16Resource('BOrd', tmpRef.bookNumber, 0, &bn); //code actually occurs earlier

		//BFCN = Book's First Chapter Number
		//err = PrvGetInt16Resource('BFCN', 0, bn-1, &bfcn);

		//clean up reference: I use number 0 to imply unspecified, as in the reference "Genesis 1" The verse is unspecified.
		if (tmpRef.chapterNumber == 0) {
			tmpRef.chapterNumber = 1;
		}
		if (tmpRef.verseNumber == 0) {
			tmpRef.verseNumber = 1;
		}

		//CFVN = Chapter's First Verse Number
		h=DmGetResource('CFVN',0);
		//make p an array of Chapters' First Verse Numbers (for entire bible, i.e. 1189 chapters for KJV)
		if (!h) return bswErrResource;
		p=(UInt16*)MemHandleLock(h);
		if (!p) {
			DmReleaseResource(h);
			return bswErrLock;
		}

		//Here is the result... Add the verse offset to the chapter's first verse number
		refP->verseRef = p[bfcn-1 + tmpRef.chapterNumber-1] + tmpRef.verseNumber-1;

		//clean up
		MemHandleUnlock(h);
		DmReleaseResource(h);
	}

    return err;
}


/************************************************************
 * Utility functions that could implemented in clients,
 * but are here to ensure they are done correctly, 
 * and to save space, as they are often used.
 */

Err bswGetTotalNumBooks(UInt16 refNum, UInt32 clientContext, UInt16 *totNumBooksP)
{
	/* Error-check our parameters */
	if( totNumBooksP == NULL) return bswErrParam;

	PrvOpenScheme(refNum, clientContext, 0);

	return PrvGetInt16Resource('TNBs', 0, 0, totNumBooksP);
}

Err bswGetBooksNumChapters(UInt16 refNum, UInt32 clientContext, UInt16 bookID, UInt16 *booksNumChaptersP)
{
	/* Error-check our parameters */
	if( booksNumChaptersP == NULL) return bswErrParam;

	PrvOpenScheme(refNum, clientContext, 0);

	if (bookID > 0) {
		return PrvGetBooksInt16Resource('BNCs', bookID, booksNumChaptersP);
	} else {
		return bswErrParam;
	}
}

Err bswGetChaptersNumVerses(UInt16 refNum, UInt32 clientContext, UInt16 bookID, UInt16 chapterNum, UInt16 *chaptersNumVersesP)
{
	/* Error-check our parameters */
	if( chaptersNumVersesP == NULL) return bswErrParam;

	PrvOpenScheme(refNum, clientContext, 0);

	if (bookID > 0) {
		if (chapterNum > 0) {
			return PrvGetChaptersInt16Resource('CNVs', bookID, chapterNum, chaptersNumVersesP);
		} else {
			return bswErrParam;
		}
	} else {
		return bswErrParam;
	}
}


/*********************************************************************
 * Private Functions
 *********************************************************************/

/************************************************************
 * FUNCTION: PrvMakeGlobals
 *
 * DESCRIPTION: Create our library globals.
 *
 * PARAMETERS:
 *
 * refNum
 *		Library reference number returned by SysLibLoad() or SysLibFind().
 *
 * CALLED BY: internal
 *
 * RETURNS:
 *
 *		pointer to our *locked* library globals
 *		NULL if our globals	could not be created.
 */

static bswGlobalsType * PrvMakeGlobals(UInt16 refNum)
{
	bswGlobalsType * gP = NULL;
	MemHandle gH;
	SysLibTblEntryType * libEntryP;

	/* Get library globals */
	libEntryP = SysLibTblEntry(refNum);
	ErrFatalDisplayIf(libEntryP == NULL, "invalid BibleSwitchLibrary refNum");

	/* Error check to make sure the globals don't already exist */
	ErrFatalDisplayIf(libEntryP->globalsP, "BibleSwitchLibrary globals already exist");

	/* Allocate and initialize our library globals. */
	gH = MemHandleNew(sizeof(bswGlobalsType));
	if ( !gH )
		return( NULL );

	/* Save the handle of our library globals in the system library table  */
	/* entry so we can later retrieve it using SysLibTblEntry(). */
	libEntryP->globalsP = (void*)gH;

	/* Lock our globals (should not fail) */
	gP = PrvLockGlobals(refNum);
	ErrFatalDisplayIf(gP == NULL, "failed to lock BibleSwitchLibrary globals");

	/* Set the owner of our globals memory chunk to "system" (zero), so it won't get
	 * freed automatically by Memory Manager when the first application to call
	 * bswOpen exits.  This is important if the library is going to stay open
	 * between apps. */
	MemPtrSetOwner(gP, 0);

	/* Initialize our library globals */
	MemSet(gP, sizeof(bswGlobalsType), 0);

	/* for convenience and debugging, save ref in globals structure */
	gP->thisLibRefNum = refNum;

	/* initial open count */
	gP->openCount = 0;

	/* return a pointer to our *locked* globals */
	return( gP );
}

/************************************************************
 * FUNCTION: PrvFreeGlobals
 *
 * DESCRIPTION: Free our library globals.
 *
 * PARAMETERS:
 *
 * refNum
 *		Library reference number returned by SysLibLoad() or SysLibFind().
 *
 * CALLED BY: internal
 *
 * RETURNS: nothing
 */

static void PrvFreeGlobals(UInt16 refNum)
{
	MemHandle gH;
	SysLibTblEntryType * libEntryP;

	/* Get our library globals handle */
	libEntryP = SysLibTblEntry(refNum);
	ErrFatalDisplayIf(libEntryP == NULL, "invalid BibleSwitchLibrary refNum");

	gH = (MemHandle)(libEntryP->globalsP);

	/* Free our library globals */
	if ( gH )
	{
		/* clear our globals reference */
		libEntryP->globalsP = NULL;

		/* free our globals */
		MemHandleFree(gH);
	}
}

/************************************************************
 * FUNCTION: PrvLockGlobals
 *
 * DESCRIPTION:	Lock our library globals
 *
 * PARAMETERS:
 *
 * refNum
 *		Library reference number returned by SysLibLoad() or SysLibFind().
 *
 * CALLED BY: internal
 *
 * RETURNS:
 *		pointer to our library globals
 *		NULL if our globals	have not been created yet.
 */

static bswGlobalsType * PrvLockGlobals(UInt16 refNum)
{
	bswGlobalsType * gP = NULL;
	MemHandle gH;
	SysLibTblEntryType * libEntryP;

	libEntryP = SysLibTblEntry(refNum);
	if ( libEntryP )
		gH = (MemHandle)(libEntryP->globalsP);
	if ( gH )
		gP = (bswGlobalsType *)MemHandleLock(gH);

	return( gP );
}

/************************************************************
 * FUNCTION: PrvIsLibOpen
 *
 * DESCRIPTION:	Check if the library has been opened.
 *
 * PARAMETERS:
 *
 * refNum
 *		Library reference number returned by SysLibLoad() or SysLibFind().
 *
 * CALLED BY: internal
 *
 * RETURNS: non-zero if the library has been opened
 */

static Boolean PrvIsLibOpen(UInt16 refNum)
{
	bswGlobalsType * gP;
	Boolean	isOpen = false;

	gP = PrvLockGlobals(refNum);

	if ( gP )
	{
		isOpen = true;
		PrvUnlockGlobals(gP);
	}

	return( isOpen );
}

/************************************************************
 * FUNCTION: PrvCreateClientContext
 *
 * DESCRIPTION:	
 *
 * Create a client context for storing client-specific data.
 * The client context allows the library to support multiple clients.
 *
 * PARAMETERS:
 *
 * gP
 *		pointer to our locked globals
 *
 * clientContextP
 *		pointer to variable for returning client context
 *
 * CALLED BY: Internal
 *
 * RETURNS:
 *		errNone
 *		memErrNotEnoughSpace
 *		bswErrNotOpen -- the library is not open
 *
 * SIDE EFFECTS:
 *
 *		*clientContextP will be set to client context on success, or zero on error.
 */

static Err PrvCreateClientContext(bswGlobalsType * gP, UInt32 * clientContextP)
{
	Err err = errNone;
	MemHandle contextH;
	bswClientContextType * contextP;
	UInt16 i;

	/* Error-check our parameters */
	ErrFatalDisplayIf(gP == NULL, "null globals pointer");
	ErrFatalDisplayIf(clientContextP == NULL, "null context variable pointer");

	/* Initialize return variable */
	*clientContextP = 0;

	/* Allocate a new client context structure */
	contextH = MemHandleNew(sizeof(bswClientContextType));
	if ( !contextH )
	{
		err = memErrNotEnoughSpace;
	}
	else
	{
		/* save context chunk handle in return variable */
		*clientContextP = (UInt32)contextH;

		/* Initialize the context chunk */
		contextP = (bswClientContextType *)MemHandleLock(contextH);

		/* save address of open routine as signature to validate context */
		contextP->libSignature = (void *)&bswOpen;

		contextP->dbRefVersification = NULL;
		contextP->schemeCode = 0;
		for (i=0;i<32;i++) {
			contextP->tListP[i].implementorID = NULL;
			contextP->tListP[i].translations = NULL;
		}

		PrvUnlockContext(contextP);

		/* increment context count (for debugging) */
		gP->contextCount++;
		ErrFatalDisplayIf(gP->contextCount == 0, "context count overflow");
	}

	return( err );
}

/************************************************************
 * FUNCTION: PrvDestroyClientContext
 *
 * DESCRIPTION:	Destroy a client context which was created by PrvCreateClientContext.
 *
 * PARAMETERS:
 *
 * gP
 *		pointer to our locked globals
 *
 * clientContextP
 *		client context
 *
 * CALLED BY: Internal
 *
 * RETURNS:
 *		errNone
 *		bswErrNotOpen -- the library is not open
 */

static Err PrvDestroyClientContext(bswGlobalsType * gP, UInt32 clientContext)
{
	bswClientContextType *	contextP;
	UInt16 i;
	UInt16 j;

	/* Error-check our parameters */
	ErrFatalDisplayIf(gP == NULL, "null globals pointer");

	/* Validate the client context by locking it */
	contextP = PrvLockContext(clientContext);

	if ( contextP )
	{
		if (contextP->dbRefVersification) DmCloseDatabase(contextP->dbRefVersification);
		contextP->schemeCode=0;
		for (i=0;i<32;i++) {
			if (contextP->tListP[i].implementorID != NULL) {
				for (j=i+1;j<32;j++) {
					if ((contextP->tListP[j].implementorID == contextP->tListP[j].implementorID) && (contextP->tListP[j].translations == contextP->tListP[i].translations)) {
						contextP->tListP[j].implementorID = NULL;
						contextP->tListP[j].translations = NULL;
					}
				}
				bswFreeTranslationList(gP->thisLibRefNum, clientContext, bswAppLaunchCmdLaunch + i, &(contextP->tListP[i].translations));
				contextP->tListP[i].translations = NULL;
			}
		}

		/* freeing a locked chunk is permitted by the system */
		MemPtrFree(contextP);

		/* decrement context count (for debugging) */
		gP->contextCount--;
		ErrFatalDisplayIf(gP->contextCount < 0, "context count underflow");
	}

	return errNone;
}

/************************************************************
 * FUNCTION: PrvLockContext
 *
 * DESCRIPTION:	Validate and lock a client context.
 *
 * PARAMETERS: 
 *
 * context
 *		a client context to lock
 *
 * CALLED BY: internal
 *
 * RETURNS: pointer to the locked client context.
 */
 
static bswClientContextType * PrvLockContext(UInt32 context)
{
	bswClientContextType *	contextP = NULL;

	/* Error-check our parameters */
	ErrFatalDisplayIf(context == 0, "null client context");

	/* Lock the client context */
	contextP = (bswClientContextType *)MemHandleLock((MemHandle)context);
	ErrFatalDisplayIf(contextP == NULL, "failed to lock client context");

	/* Validate the client context */
	ErrFatalDisplayIf(contextP->libSignature != (void *)&bswOpen,
		"invalid client context");

	return( contextP );
}


/*********************************************************************
 * Private Utility Functions
 *********************************************************************/

static Err PrvOpenScheme(UInt16 refNum, UInt32 clientContext, UInt32 schemeCode)
{
	bswClientContextTypePtr contextP;
	DmSearchStateType state;
	UInt16 card;
	LocalID currentDB = 0;
	Err err=errNone;
	Boolean newSearch = true;
	Boolean opened = false;
	MemHandle h;
	UInt32* p;
	
	if ( !PrvIsLibOpen(refNum) )
		return( bswErrNotOpen );

	contextP = PrvLockContext(clientContext);
	
	if ( (schemeCode != 0) && (schemeCode != contextP->schemeCode) ) {
		if (contextP->dbRefVersification) {
			DmCloseDatabase(contextP->dbRefVersification);
			contextP->dbRefVersification = NULL;
		}
		contextP->schemeCode=0;
	}

	while ( (err == errNone) && ( contextP->dbRefVersification == NULL) ) {
		err=DmGetNextDatabaseByTypeCreator(newSearch, &state, bswVersificationType, bswCreatorID, false, &card, &currentDB);
		newSearch=false;
		if ( err == errNone ) {
			//stop when resource VRef == schemeCode or schemeCode == 0
			contextP->dbRefVersification = DmOpenDatabase(card, currentDB, dmModeReadOnly);
			if (contextP->dbRefVersification) {
				if (schemeCode == 0) {
					opened=true;
				} else {
					h = DmGetResource('VRef',0);
					if (h) {
						p = (UInt32*)MemHandleLock(h);
						if (p) {
							if (*p == schemeCode) {
								opened=true;
								contextP->schemeCode=schemeCode;
							}
							MemHandleUnlock(h);
						}
						DmReleaseResource(h);
					}
				}
			}
			if (!opened) {
				DmCloseDatabase(contextP->dbRefVersification);
				contextP->dbRefVersification = NULL;
			}
		}
	}

	//TODO: Handle this better....
	ErrFatalDisplayIf(contextP->dbRefVersification == NULL, "No Versifications!");
	
	PrvUnlockContext(contextP);
	
	if (opened) {
		return errNone;
	} else {
		return( bswErrResource );
	}
}

/************************************************************
 * FUNCTION: PrvImplmentationCall
 *
 * DESCRIPTION:
 *
 * ?
 * 
 * PARAMETERS:
 *
 * ?
 *
 * CALLED BY: internal
 *
 * RETURNS:
 *		errNone
 */

static Err PrvImplementationCall(UInt16 refNum, MemPtr cmdPBP, UInt16 launchCode)
{
	UInt16 cardNo;
	LocalID dbID;
	Err err = errNone;
	UInt32 result;
	
	err = PrvGetImplementation(refNum, launchCode, &cardNo, &dbID);
	if( errNone == err ) {
		err = SysAppLaunch(cardNo, dbID, 0, launchCode, cmdPBP, &result);
		if (errNone == err) {
			err = result;
		}
	}

    return err;
}    

/************************************************************
 * FUNCTION: PrvGetImplementation
 *
 * DESCRIPTION:
 *
 * ?
 * 
 * PARAMETERS:
 *
 * ?
 *
 * CALLED BY: internal
 *
 * RETURNS:
 *		errNone
 */

static Err PrvGetImplementation(UInt16 refNum, bswLaunchCode launchCode, UInt16 *cardNoP, LocalID *dbIDP)
{
	DmSearchStateType stateInfo;
	UInt32 implementingCreatorCode;
	UInt16 prefsSize;
	Err err = errNone;
	Int16 prefsVersion;
	UInt16 bswCallNumber;
	
	/* Error-check our parameters */
	if( launchCode == NULL) return bswErrParam;

	prefsSize=4;
	implementingCreatorCode=0;

	switch (launchCode) {
	case bswAppLaunchCmdLaunch:
		bswCallNumber = 1;
		break;
	case bswAppLaunchCmdGetReference:
		bswCallNumber = 2;
		break;
	case bswAppLaunchCmdFindText:
		bswCallNumber = 3;
		break;
	case bswAppLaunchCmdSelectReference:
		bswCallNumber = 4;
		break;
	case bswAppLaunchCmdSelectedReference:
		bswCallNumber = 1;
		break;
	case bswAppLaunchCmdGetCount:
		bswCallNumber = 1;
		break;
	default:
		bswCallNumber = 0x7FFF; // Unknown launch code... wrong version of library?
	}

	prefsVersion=PrefGetAppPreferences(bswPreferencesCreatorID,bswCallNumber,&implementingCreatorCode,&prefsSize,true);
	if( ( prefsVersion == noPreferenceFound) || (implementingCreatorCode == 0 ) ) {
		err = dmErrCantFind;
	} else {
		err = DmGetNextDatabaseByTypeCreator(true, &stateInfo, 'appl', implementingCreatorCode, true, cardNoP, dbIDP);
	}

    return err;
}

/************************************************************
 * FUNCTION: PrvGetBooksInt16Resource
 *
 * DESCRIPTION:
 *
 * ?
 * 
 * PARAMETERS:
 *
 * ?
 *
 * CALLED BY: internal
 *
 * RETURNS:
 *		errNone
 */

static Err PrvGetBooksInt16Resource(UInt32 recordType, UInt16 bookID, UInt16 *resultP)
{
	UInt16 seqNum;
	Err err;
	err = PrvGetInt16Resource('BOrd', bookID, 0, &seqNum);
	if ( err == errNone ) {
		return PrvGetInt16Resource(recordType, 0, seqNum-1, resultP);
	} else {
		return err;
	}
}

/************************************************************
 * FUNCTION: PrvGetChaptersInt16Resource
 *
 * DESCRIPTION:
 *
 * ?
 * 
 * PARAMETERS:
 *
 * ?
 *
 * CALLED BY: internal
 *
 * RETURNS:
 *		errNone
 */

static Err PrvGetChaptersInt16Resource(UInt32 recordType, UInt16 bookID, UInt16 chapterNum, UInt16 *resultP)
{
	UInt16 seqNum;
	Err err;

	if (bookID > 0) {
		err = PrvGetBooksInt16Resource('BFCN', bookID, &seqNum);
	} else {
		/* Unspecified book, use first chapter of bible. */
		err = PrvGetInt16Resource('BFCN', 0, 0, &seqNum);
	}

	if ( err == errNone ) {
		return PrvGetInt16Resource(recordType, 0, seqNum-1 + chapterNum-1, resultP);
	} else {
		return err;
	}
}

/************************************************************
 * FUNCTION: PrvGetInt16Resource
 *
 * DESCRIPTION:
 *
 * ?
 * 
 * PARAMETERS:
 *
 * ?
 *
 * CALLED BY: internal
 *
 * RETURNS:
 *		errNone
 */

static Err PrvGetInt16Resource(UInt32 recordType, UInt16 recordID, UInt16 elementNumber, UInt16 *resultP)
{
	MemHandle h = NULL;
	UInt16* p = NULL;
	Err err=errNone;

	h=DmGetResource(recordType,recordID);
	if (!h) {
		return bswErrResource;
	}
	p=(UInt16*)MemHandleLock(h);
	if (!p) {
	    DmReleaseResource(h);
		return bswErrLock;
	}

	if (elementNumber <= (MemHandleSize(h)>>1)){
		*resultP = p[elementNumber];
	} else {
		err=bswErrParam;
	}

    MemHandleUnlock(h);
    DmReleaseResource(h);
    return err;
}
